/* This file is part of flowz.
   Copyright (C) 2006 Sergey Poznyakoff

   Flowz is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Flowz is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with flowz; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
   MA 02110-1301 USA */

#include "flowz.h"
#define obstack_chunk_alloc malloc
#define obstack_chunk_free free

static Hash_table *flow_table;

struct table_entry {
     time_t time_start;
     time_t time_end;
     uintmax_t octets;
     char key[1];
};

/* Calculate the hash of a string.  */
static size_t
flow_table_hasher(void const *data, unsigned n_buckets)
{
     struct table_entry const *t = data;
     return hash_string(t->key, n_buckets);
}

/* Compare two strings for equality.  */
static bool
flow_table_compare(void const *data1, void const *data2)
{
     struct table_entry const *t1 = data1;
     struct table_entry const *t2 = data2;
     return strcmp(t1->key, t2->key) == 0;
}

static struct obstack stk;
static void *mem_top;

void
table_init()
{
     flow_table = hash_initialize(0, 0, 
				  flow_table_hasher,
				  flow_table_compare,
				  NULL);
     if (!flow_table)
	  xalloc_die();
     obstack_init(&stk);
     mem_top = obstack_alloc(&stk, 1);
}

struct table_entry *
lookup_or_install(struct table_entry *tab)
{
     struct table_entry *ret = hash_insert(flow_table, tab);
     if (!ret)
	  xalloc_die();
     return ret;
}

size_t
table_n_entries()
{
     return hash_get_n_entries(flow_table);
}

int
update(char *key, time_t time_start, time_t time_end, uintmax_t octets)
{
     struct table_entry *tab, *tp;
     
     tab = obstack_alloc(&stk, sizeof(*tab) + strlen(key));
     memset(tab, 0, sizeof *tab);
     strcpy(tab->key, key);
     tp = lookup_or_install(tab);
     if (tp == tab) 
	  tab = NULL;
     else
	  obstack_free(&stk, tab);
     if (tp->time_start == 0 || tp->time_start > time_start)
	  tp->time_start = time_start;
     if (tp->time_end < time_end)
	  tp->time_end = time_end;

     if (TYPE_MAXIMUM(off_t) - tp->octets < octets) {
	  error(0, 0, "%s:%lu: integer overflow",
		input, (unsigned long) lineno);
	  return 1;
     }
     tp->octets += octets;

     if (max_memory_usage) {
	  off_t estimate =
	       obstack_memory_used(&stk) +
	       hash_get_n_entries(flow_table) * sizeof (struct hash_entry);
	  if (estimate >= max_memory_usage) {
	       dump_table(outfile, header_str);
	       hash_clear(flow_table);
	       obstack_free(&stk, mem_top);
	       header_str = NULL; /* make sure header is not printed twice */
	  }
     }
     
     return 0;
}

static bool
dumper(void *data, void *proc_data)
{
     struct table_entry *tp = data;
     FILE *fp = proc_data;
     char buf[UINTMAX_STRSIZE_BOUND];

     fprintf(fp, "%s-", umaxtostr(tp->time_start, buf));
     fprintf(fp, "%s,", umaxtostr(tp->time_end, buf));
     fprintf(fp, "%s,%s\n", umaxtostr(tp->octets, buf), tp->key);
     return true;
}

void
dump_table(FILE *fp, const char *header)
{
     if (flow_table) {
	  if (header)
	       fprintf(fp, "#:%s\n", header);
	  hash_do_for_each(flow_table, dumper, fp);
     }
}

void
table_stat(FILE *fp)
{
     hash_print_statistics(flow_table, fp);
}
